"""Build beta detail models for the API"""

import enum

import deserialize

from asconnect.models.common import BaseAttributes, Links, Relationship, Resource


class ExternalBetaState(enum.Enum):
    """External beta state."""

    PROCESSING = "PROCESSING"
    PROCESSING_EXCEPTION = "PROCESSING_EXCEPTION"
    MISSING_EXPORT_COMPLIANCE = "MISSING_EXPORT_COMPLIANCE"
    READY_FOR_BETA_TESTING = "READY_FOR_BETA_TESTING"
    IN_BETA_TESTING = "IN_BETA_TESTING"
    EXPIRED = "EXPIRED"
    READY_FOR_BETA_SUBMISSION = "READY_FOR_BETA_SUBMISSION"
    IN_EXPORT_COMPLIANCE_REVIEW = "IN_EXPORT_COMPLIANCE_REVIEW"
    WAITING_FOR_BETA_REVIEW = "WAITING_FOR_BETA_REVIEW"
    IN_BETA_REVIEW = "IN_BETA_REVIEW"
    BETA_REJECTED = "BETA_REJECTED"
    BETA_APPROVED = "BETA_APPROVED"


class InternalBetaState(enum.Enum):
    """Internal beta state."""

    PROCESSING = "PROCESSING"
    PROCESSING_EXCEPTION = "PROCESSING_EXCEPTION"
    MISSING_EXPORT_COMPLIANCE = "MISSING_EXPORT_COMPLIANCE"
    READY_FOR_BETA_TESTING = "READY_FOR_BETA_TESTING"
    IN_BETA_TESTING = "IN_BETA_TESTING"
    EXPIRED = "EXPIRED"
    IN_EXPORT_COMPLIANCE_REVIEW = "IN_EXPORT_COMPLIANCE_REVIEW"


@deserialize.key("identifier", "id")
class BuildBetaDetail(Resource):
    """Represents a build localization."""

    @deserialize.key("auto_notify_enabled", "autoNotifyEnabled")
    @deserialize.key("external_build_state", "externalBuildState")
    @deserialize.key("internal_build_state", "internalBuildState")
    class Attributes(BaseAttributes):
        """Represents beta build localization attributes."""

        auto_notify_enabled: bool
        external_build_state: ExternalBetaState
        internal_build_state: InternalBetaState

    identifier: str
    attributes: Attributes
    relationships: dict[str, Relationship] | None
    links: Links



"""App Models for the API"""

import enum

import deserialize

from asconnect.models.common import BaseAttributes, Resource, Links, Relationship, Reprable


class ScreenshotDisplayType(enum.Enum):
    """Screenshot display type."""

    APP_IPHONE_67 = "APP_IPHONE_67"
    APP_IPHONE_65 = "APP_IPHONE_65"
    APP_IPHONE_61 = "APP_IPHONE_61"
    APP_IPHONE_58 = "APP_IPHONE_58"
    APP_IPHONE_55 = "APP_IPHONE_55"
    APP_IPHONE_47 = "APP_IPHONE_47"
    APP_IPHONE_40 = "APP_IPHONE_40"
    APP_IPHONE_35 = "APP_IPHONE_35"
    APP_IPAD_PRO_3GEN_129 = "APP_IPAD_PRO_3GEN_129"
    APP_IPAD_PRO_3GEN_11 = "APP_IPAD_PRO_3GEN_11"
    APP_IPAD_PRO_129 = "APP_IPAD_PRO_129"
    APP_IPAD_105 = "APP_IPAD_105"
    APP_IPAD_97 = "APP_IPAD_97"
    APP_WATCH_ULTRA = "APP_WATCH_ULTRA"
    APP_WATCH_SERIES_7 = "APP_WATCH_SERIES_7"
    APP_WATCH_SERIES_4 = "APP_WATCH_SERIES_4"
    APP_WATCH_SERIES_3 = "APP_WATCH_SERIES_3"
    APP_DESKTOP = "APP_DESKTOP"
    APP_APPLE_TV = "APP_APPLE_TV"
    IMESSAGE_APP_IPHONE_67 = "IMESSAGE_APP_IPHONE_67"
    IMESSAGE_APP_IPHONE_65 = "IMESSAGE_APP_IPHONE_65"
    IMESSAGE_APP_IPHONE_61 = "IMESSAGE_APP_IPHONE_61"
    IMESSAGE_APP_IPHONE_58 = "IMESSAGE_APP_IPHONE_58"
    IMESSAGE_APP_IPHONE_55 = "IMESSAGE_APP_IPHONE_55"
    IMESSAGE_APP_IPHONE_47 = "IMESSAGE_APP_IPHONE_47"
    IMESSAGE_APP_IPHONE_40 = "IMESSAGE_APP_IPHONE_40"
    IMESSAGE_APP_IPAD_PRO_3GEN_129 = "IMESSAGE_APP_IPAD_PRO_3GEN_129"
    IMESSAGE_APP_IPAD_PRO_3GEN_11 = "IMESSAGE_APP_IPAD_PRO_3GEN_11"
    IMESSAGE_APP_IPAD_PRO_129 = "IMESSAGE_APP_IPAD_PRO_129"
    IMESSAGE_APP_IPAD_105 = "IMESSAGE_APP_IPAD_105"
    IMESSAGE_APP_IPAD_97 = "IMESSAGE_APP_IPAD_97"

    @staticmethod
    def from_name(name: str) -> "ScreenshotDisplayType":
        """Generate the display type from an image name.

        :param name: The name to convert

        :returns: The corresponding display type
        """
        identifier = name.split("-")[0]
        return ScreenshotDisplayType(identifier)


@deserialize.key("identifier", "id")
class AppScreenshotSet(Resource):
    """Represents an app store screenshot set."""

    @deserialize.key("screenshot_display_type", "screenshotDisplayType")
    class Attributes(BaseAttributes):
        """Attributes."""

        screenshot_display_type: ScreenshotDisplayType

    identifier: str
    attributes: Attributes
    relationships: dict[str, Relationship] | None
    links: Links


class AppMediaStateError(Reprable):
    """An app media state error."""

    code: str
    description: str


class AppMediaAssetStateState(enum.Enum):
    """The state value for and app media asset state."""

    AWAITING_UPLOAD = "AWAITING_UPLOAD"
    UPLOAD_COMPLETE = "UPLOAD_COMPLETE"
    COMPLETE = "COMPLETE"
    FAILED = "FAILED"


class AppMediaAssetState(Reprable):
    """An app media asset state."""

    errors: list[AppMediaStateError]
    state: AppMediaAssetStateState
    warnings: list[AppMediaStateError] | None


@deserialize.key("template_url", "templateUrl")
class ImageAsset(Reprable):
    """An image asset."""

    template_url: str
    height: int
    width: int


class UploadOperationHeader(Reprable):
    """An upload operation header."""

    name: str
    value: str


@deserialize.key("request_headers", "requestHeaders")
class UploadOperation(Reprable):
    """An upload operation."""

    length: int
    method: str
    offset: int
    request_headers: list[UploadOperationHeader]
    url: str


@deserialize.key("identifier", "id")
class AppScreenshot(Resource):
    """Represents an app store screenshot."""

    @deserialize.key("asset_delivery_state", "assetDeliveryState")
    @deserialize.key("asset_token", "assetToken")
    @deserialize.key("asset_type", "assetType")
    @deserialize.key("file_name", "fileName")
    @deserialize.key("file_size", "fileSize")
    @deserialize.key("image_asset", "imageAsset")
    @deserialize.key("source_file_checksum", "sourceFileChecksum")
    @deserialize.key("upload_operations", "uploadOperations")
    class Attributes(BaseAttributes):
        """Attributes."""

        asset_delivery_state: AppMediaAssetState
        asset_token: str
        asset_type: str
        file_name: str
        file_size: int | None
        image_asset: ImageAsset | None
        source_file_checksum: str | None
        uploaded: bool | None
        upload_operations: list[UploadOperation] | None

    identifier: str
    attributes: Attributes
    relationships: dict[str, Relationship] | None
    links: Links



"""Model types for requests."""

import deserialize

from asconnect.models.common import BaseAttributes, Resource, Links, Relationship


@deserialize.key("identifier", "id")
class BetaAppReviewDetail(Resource):
    """Represents a beta apps review details."""

    @deserialize.key("contact_email", "contactEmail")
    @deserialize.key("contact_first_name", "contactFirstName")
    @deserialize.key("contact_last_name", "contactLastName")
    @deserialize.key("contact_phone", "contactPhone")
    @deserialize.key("demo_account_name", "demoAccountName")
    @deserialize.key("demo_account_password", "demoAccountPassword")
    @deserialize.key("demo_account_required", "demoAccountRequired")
    class Attributes(BaseAttributes):
        """Attributes."""

        contact_email: str
        contact_first_name: str
        contact_last_name: str
        contact_phone: str
        demo_account_name: str | None
        demo_account_password: str | None
        demo_account_required: bool
        notes: str | None

    identifier: str
    attributes: Attributes
    relationships: dict[str, Relationship] | None
    links: Links



"""Wrapper around the Apple App Store Connect APIs."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
from typing import Any, Iterator

from asconnect.httpclient import HttpClient

from asconnect.models import (
    BetaAppLocalization,
    BetaAppReviewDetail,
    BetaBuildLocalization,
    BetaGroup,
)


class BetaReviewClient:
    """Wrapper class around the ASC API."""

    log: logging.Logger
    http_client: HttpClient

    def __init__(
        self,
        *,
        http_client: HttpClient,
        log: logging.Logger,
    ) -> None:
        """Construct a new client object.

        :param http_client: The API HTTP client
        :param log: Any base logger to be used (one will be created if not supplied)
        """

        self.http_client = http_client
        self.log = log.getChild("beta_review")

    # pylint:disable=too-many-arguments
    def set_beta_app_review_details(
        self,
        *,
        app_id: str,
        contact_email: str,
        contact_first_name: str,
        contact_last_name: str,
        contact_phone: str,
        demo_account_name: str | None = None,
        demo_account_password: str | None = None,
        demo_account_required: bool | None = None,
        notes: str | None = None,
    ) -> BetaAppReviewDetail:
        """Set the Beta app review details.

        :param app_id: The Apple ID for the app
        :param contact_email: The email for the app review contact
        :param contact_first_name: The first name for the app review contact
        :param contact_last_name: The last name for the app review contact
        :param contact_phone: The phone number for the app review contact
        :param demo_account_name: The username for the demo account
        :param demo_account_password: The password for the demo account
        :param demo_account_required: Set to True to mark the demo account as required
        :param notes: Any notes for the reviewer

        :returns: The raw response
        """

        self.log.info(f"Setting beta app review details on {app_id}")

        attributes: dict[str, Any] = {
            "contactEmail": contact_email,
            "contactFirstName": contact_first_name,
            "contactLastName": contact_last_name,
            "contactPhone": contact_phone,
        }

        if demo_account_name is not None:
            attributes["demoAccountName"] = demo_account_name

        if demo_account_password is not None:
            attributes["demoAccountPassword"] = demo_account_password

        if demo_account_required is not None:
            attributes["demoAccountRequired"] = demo_account_required

        if notes is not None:
            attributes["notes"] = notes

        body = {
            "data": {
                "attributes": attributes,
                "id": app_id,
                "type": "betaAppReviewDetails",
            }
        }

        self.log.debug(f"Beta review details: {body}")

        return self.http_client.patch(
            endpoint=f"betaAppReviewDetails/{app_id}",
            data=body,
            data_type=BetaAppReviewDetail,
        )

    # pylint:enable=too-many-arguments

    def get_beta_app_localizations(self, app_id: str) -> Iterator[BetaAppLocalization]:
        """Get the beta app localizations.

        :param app_id: The apple identifier for the app to get the localizations for

        :returns: An iterator to the beta app localizations
        """
        self.log.debug(f"Getting beta app localizations for {app_id}")
        url = self.http_client.generate_url(f"apps/{app_id}/betaAppLocalizations")
        yield from self.http_client.get(url=url, data_type=list[BetaAppLocalization])

    def get_beta_build_localizations(self, build_id: str) -> Iterator[BetaBuildLocalization]:
        """Get the beta app localizations.

        :param build_id: The identifier for the build to get the localizations for

        :returns: An iterator to the beta app localizations
        """
        self.log.debug(f"Getting beta build localizations for {build_id}")
        url = self.http_client.generate_url(f"betaBuildLocalizations?filter[build]={build_id}")
        yield from self.http_client.get(url=url, data_type=list[BetaBuildLocalization])

    def set_beta_app_localizations(
        self, app_id: str, localizations: dict[str, dict[str, str]]
    ) -> None:
        """Set the app localizations.

        :param app_id: The apple identifier for the app to set the localizations for
        :param localizations: A dictionary of language codes to localization attributes
        """

        self.log.info(f"Setting beta app localizations for {app_id}")

        existing_localizations = {}

        for localization in self.get_beta_app_localizations(app_id):
            existing_localizations[localization.attributes.locale] = localization

        for language_code, language_info in localizations.items():
            self.log.debug(f"Setting localizations for {language_code}: {language_info}")
            existing_localization = existing_localizations.get(language_code)

            self.log.debug(f"Existing localization: {existing_localization}")

            if existing_localization:
                self.http_client.patch(
                    endpoint=f"betaAppLocalizations/{existing_localization.identifier}",
                    data={
                        "data": {
                            "attributes": language_info,
                            "id": existing_localization.identifier,
                            "type": "betaAppLocalizations",
                        }
                    },
                    data_type=BetaAppLocalization,
                )
            else:
                language_info["locale"] = language_code
                self.http_client.post(
                    endpoint="betaAppLocalizations",
                    data={
                        "data": {
                            "attributes": language_info,
                            "type": "betaAppLocalizations",
                            "relationships": {
                                "app": {
                                    "data": {
                                        "type": "apps",
                                        "id": app_id,
                                    }
                                }
                            },
                        }
                    },
                )

    def set_whats_new_for_build(self, build_id: str, localizations: dict[str, str]) -> None:
        """Set the whats new for a build.

        :param build_id: The apple identifier for the app to set the localizations for
        :param localizations: A dictionary of language codes to localization info
        """

        self.log.info(f"Setting whats new for build {build_id}")
        self.log.debug(f"Localizations: {localizations}")

        existing_localizations = {}

        for localization in self.get_beta_build_localizations(build_id):
            existing_localizations[localization.attributes.locale] = localization

        self.log.debug(f"Existing localizations: {existing_localizations}")

        for language_code, whats_new in localizations.items():
            self.log.debug(f"Setting localization for {language_code}")
            attributes = {"whatsNew": whats_new}

            existing_localization = existing_localizations.get(language_code)

            if existing_localization:
                self.http_client.patch(
                    endpoint=f"betaBuildLocalizations/{existing_localization.identifier}",
                    data={
                        "data": {
                            "attributes": attributes,
                            "id": existing_localization.identifier,
                            "type": "betaBuildLocalizations",
                        }
                    },
                    data_type=BetaBuildLocalization,
                )
            else:
                attributes["locale"] = language_code
                self.http_client.post(
                    endpoint="betaBuildLocalizations",
                    data={
                        "data": {
                            "attributes": attributes,
                            "type": "betaBuildLocalizations",
                            "relationships": {
                                "build": {
                                    "data": {
                                        "type": "builds",
                                        "id": build_id,
                                    }
                                }
                            },
                        }
                    },
                )

    def get_beta_groups(self, app_id: str) -> Iterator[BetaGroup]:
        """Get the beta groups

        :param app_id: The ID of the app to filter on

        :returns: An iterator to the beta groups
        """
        self.log.debug(f"Getting beta groups for {app_id}")
        url = self.http_client.generate_url(f"betaGroups?filter[app]={app_id}")
        yield from self.http_client.get(url=url, data_type=list[BetaGroup])

    def set_beta_groups_on_build(self, build_id: str, beta_groups: list[BetaGroup]) -> None:
        """Set the Beta groups on a build.

        :param build_id: The build ID for the build to set the groups on
        :param beta_groups: The groups to add
        """
        self.log.info(f"Setting beta groups on {build_id}: {beta_groups}")

        data = []
        for beta_group in beta_groups:
            data.append({"type": "betaGroups", "id": beta_group.identifier})

        body = {"data": data}

        self.http_client.post(endpoint=f"builds/{build_id}/relationships/betaGroups", data=body)

    def submit_for_beta_review(self, build_id: str) -> None:
        """Submit a build for beta review

        :param build_id: The build ID for the build to set the groups on
        """

        self.log.info(f"Submitting build for beta review: {build_id}")

        body = {
            "data": {
                "type": "betaAppReviewSubmissions",
                "relationships": {
                    "build": {
                        "data": {
                            "id": build_id,
                            "type": "builds",
                        }
                    }
                },
            }
        }

        self.log.debug(f"Data: {body}")

        self.http_client.post(endpoint="betaAppReviewSubmissions", data=body)



"""User Models for the API"""

import enum

import deserialize

from asconnect.models.common import BaseAttributes, Links, Relationship, Resource


class UserRole(enum.Enum):
    """Represents a user role."""

    ACCESS_TO_REPORTS = "ACCESS_TO_REPORTS"
    ACCOUNT_HOLDER = "ACCOUNT_HOLDER"
    ADMIN = "ADMIN"
    APP_MANAGER = "APP_MANAGER"
    CLOUD_MANAGED_APP_DISTRIBUTION = "CLOUD_MANAGED_APP_DISTRIBUTION"
    CLOUD_MANAGED_DEVELOPER_ID = "CLOUD_MANAGED_DEVELOPER_ID"
    CREATE_APPS = "CREATE_APPS"
    CUSTOMER_SUPPORT = "CUSTOMER_SUPPORT"
    DEVELOPER = "DEVELOPER"
    FINANCE = "FINANCE"
    IMAGE_MANAGER = "IMAGE_MANAGER"
    MARKETING = "MARKETING"
    SALES = "SALES"


@deserialize.key("first_name", "firstName")
@deserialize.key("last_name", "lastName")
@deserialize.key("provisioning_allowed", "provisioningAllowed")
@deserialize.key("all_apps_visible", "allAppsVisible")
class UserAttributes(BaseAttributes):
    """Represents build attributes."""

    first_name: str
    last_name: str
    roles: list[UserRole]
    provisioning_allowed: bool
    all_apps_visible: bool
    username: str


@deserialize.key("identifier", "id")
class User(Resource):
    """Represents a user."""

    identifier: str
    attributes: UserAttributes
    relationships: dict[str, Relationship] | None
    links: Links



"""Utilities for the library."""

import hashlib
import os
from typing import Iterator, TypeVar
import urllib.parse

IteratorType = TypeVar("IteratorType")  # pylint: disable=invalid-name


def next_or_none(iterator: Iterator[IteratorType]) -> IteratorType | None:
    """Get the next value from an iterator, or return None when it is exhausted.

    :param iterator: The iterator to get the next value from

    :returns: The next value or None if exhausted
    """
    try:
        return next(iterator)
    except StopIteration:
        return None


def update_query_parameters(url: str, query_parameters: dict[str, str]) -> str:
    """Update the query parameters on a URL.

    :param url: The URL to update
    :param query_parameters: The query parameters to add

    :returns: The updated URL
    """
    parsed_url = urllib.parse.urlparse(url)
    parsed_parameters = dict(urllib.parse.parse_qsl(parsed_url.query))

    new_parameters = {**parsed_parameters, **query_parameters}
    new_parameter_string = urllib.parse.urlencode(new_parameters, safe="[]")

    parsed_url = urllib.parse.ParseResult(**dict(parsed_url._asdict(), query=new_parameter_string))

    return urllib.parse.urlunparse(parsed_url)


def md5_file(file_path: str) -> str:
    """Generate the MD5 of a file.

    :param file_path: The file to generate the MD5 for

    :returns: The MD5 as a hex string
    """
    hasher = hashlib.md5()
    with open(file_path, "rb") as file_handle:
        for chunk in iter(lambda: file_handle.read(4096), b""):
            hasher.update(chunk)
    return hasher.hexdigest()


def write_key(key_id: str, key_contents: str) -> str:
    """Write a key to the private key folder for altool.

    :param key_id: The ID of the key
    :param key_contents: The text key contents

    :returns: The path the key was written out to
    """

    folder_path = os.path.expanduser("~/.appstoreconnect/private_keys")
    os.makedirs(folder_path, exist_ok=True)

    key_file_name = f"AuthKey_{key_id}.p8"
    key_file_path = os.path.join(folder_path, key_file_name)

    with open(key_file_path, "w", encoding="utf-8") as key_file:
        key_file.write(key_contents)

    return key_file_path



"""App Models for the API"""

import enum

import deserialize

from asconnect.models.common import BaseAttributes, Resource, Links, Relationship


class ContentRightsDeclaration(enum.Enum):
    """Contents rights declarations."""

    DOES_NOT_USE_THIRD_PARTY_CONTENT = "DOES_NOT_USE_THIRD_PARTY_CONTENT"
    USES_THIRD_PARTY_CONTENT = "USES_THIRD_PARTY_CONTENT"


@deserialize.auto_snake()
class AppAttributes(BaseAttributes):
    """Represents app attributes."""

    bundle_id: str
    name: str
    primary_locale: str
    sku: str
    available_in_new_territories: bool | None
    content_rights_declaration: ContentRightsDeclaration | None
    is_or_ever_was_made_for_kids: bool
    subscription_status_url: str | None
    subscription_status_url_version: str | None
    subscription_status_url_for_sandbox: str | None
    subscription_status_url_version_for_sandbox: str | None
    streamlined_purchasing_enabled: bool | None


@deserialize.key("identifier", "id")
class App(Resource):
    """Represents an app."""

    identifier: str
    attributes: AppAttributes
    relationships: dict[str, Relationship] | None
    links: Links

    @property
    def bundle_id(self) -> str:
        """Return the bundle ID.

        :returns: The bundle ID
        """
        return self.attributes.bundle_id

    @property
    def name(self) -> str:
        """Return the name

        :returns: The name
        """
        return self.attributes.name

    @property
    def primary_locale(self) -> str:
        """Return the primary locale

        :returns: The primary locale
        """
        return self.attributes.primary_locale

    @property
    def sku(self) -> str:
        """Return the SKU

        :returns: The SKU
        """
        return self.attributes.sku

    @property
    def available_in_new_territories(self) -> bool | None:
        """Returns whether or not this app is available in new territories

        :returns: True if this app is available in new territories, False otherwise
        """
        return self.attributes.available_in_new_territories

    @property
    def content_rights_declaration(self) -> ContentRightsDeclaration | None:
        """Return any content rights declaration

        :returns: Any content rights declaration
        """
        return self.attributes.content_rights_declaration

    @property
    def is_or_ever_was_made_for_kids(self) -> bool:
        """Return whether or not this app was ever made for kids

        :returns: True if this app was ever made for kids, False otherwise
        """
        return self.attributes.is_or_ever_was_made_for_kids



"""Localization models for the API"""

import deserialize

from asconnect.models.common import BaseAttributes, Links, Relationship, Resource


@deserialize.key("identifier", "id")
class BetaAppLocalization(Resource):
    """Represents a build."""

    @deserialize.key("feedback_email", "feedbackEmail")
    @deserialize.key("marketing_url", "marketingUrl")
    @deserialize.key("privacy_policy_url", "privacyPolicyUrl")
    @deserialize.key("tv_os_privacy_policy", "tvOsPrivacyPolicy")
    class Attributes(BaseAttributes):
        """Represents beta app localization attributes."""

        description: str
        feedback_email: str
        locale: str
        marketing_url: str | None
        privacy_policy_url: str | None
        tv_os_privacy_policy: str | None

    identifier: str
    attributes: Attributes
    relationships: dict[str, Relationship] | None
    links: Links


@deserialize.key("identifier", "id")
class BetaBuildLocalization(Resource):
    """Represents a build localization."""

    @deserialize.key("whats_new", "whatsNew")
    class Attributes(BaseAttributes):
        """Represents beta build localization attributes."""

        locale: str
        whats_new: str | None

    identifier: str
    attributes: Attributes
    relationships: dict[str, Relationship] | None
    links: Links



"""Wrapper around the Apple App Store Connect APIs."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
from typing import Iterator

from asconnect.httpclient import HttpClient

from asconnect.models import CustomerReview
from asconnect.sorting import CustomerReviewSort
from asconnect.utilities import update_query_parameters


class ReviewsClient:
    """Wrapper class around the ASC API."""

    log: logging.Logger
    http_client: HttpClient

    def __init__(
        self,
        *,
        http_client: HttpClient,
        log: logging.Logger,
    ) -> None:
        """Construct a new client object.

        :param http_client: The API HTTP client
        :param log: Any base logger to be used (one will be created if not supplied)
        """

        self.http_client = http_client
        self.log = log.getChild("reviews")

    def get_reviews(
        self,
        app_id: str,
        sort_order: CustomerReviewSort | None = None,
        territory_filter: list[str] | None = None,
        published_response: bool | None = None,
    ) -> Iterator[CustomerReview]:
        """Get customer reviews for an app.

        :param app_id: The app ID to get reviews for
        :param sort_order: The order to sort the reviews in. Defaults to None.
        :param territory_filter: The territory to filter the reviews by. Defaults to None.
        :param published_response: If set to True, only reviews with a published response will be
                                   returned. If set to False, only reviews without a published
                                   response will be returned. Defaults to None which returns all.

        :yields: An iterator of reviews
        """

        self.log.info("Getting users...")

        url = self.http_client.generate_url(f"apps/{app_id}/customerReviews")

        query_parameters = {}

        if sort_order:
            query_parameters["sort"] = sort_order.value

        if territory_filter:
            if isinstance(territory_filter, str):
                territory_filter = [territory_filter]
            query_parameters["filter[territory]"] = ",".join(territory_filter)

        if published_response is not None:
            query_parameters["exists[publishedResponse]"] = str(published_response).lower()

        url = update_query_parameters(url, query_parameters)

        yield from self.http_client.get(url=url, data_type=list[CustomerReview])



"""Wrapper around the Apple App Store Connect APIs."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
from typing import Any, Iterator

from asconnect.httpclient import HttpClient
from asconnect.models import App, AppStoreVersion, Platform, ReleaseType


class AppClient:
    """Wrapper class around the ASC API."""

    log: logging.Logger
    http_client: HttpClient

    def __init__(
        self,
        *,
        http_client: HttpClient,
        log: logging.Logger,
    ) -> None:
        """Construct a new client object.

        :param http_client: The API HTTP client
        :param log: Any base logger to be used (one will be created if not supplied)
        """

        self.http_client = http_client
        self.log = log.getChild("app")

    def get_all(
        self,
        url: str | None = None,
    ) -> Iterator[App]:
        """Get all apps.

        :param url: The URL to use (will be generated if not supplied)

        :returns: A list of apps
        """
        self.log.debug("Getting all apps...")
        url = self.http_client.generate_url("apps")
        yield from self.http_client.get(url=url, data_type=list[App])

    def get_from_bundle_id(self, bundle_id: str) -> App | None:
        """Get a particular app.

        :param bundle_id: The bundle ID of the app to get

        :returns: The app if found, None otherwise
        """
        self.log.debug(f"Getting app with bundle id '{bundle_id}'...")
        for app in self.get_all():
            if app.bundle_id == bundle_id:
                return app
        return None

    def create_new_version(
        self,
        *,
        version: str,
        app_id: str,
        platform: Platform = Platform.IOS,
        copyright_text: str | None = None,
        uses_idfa: bool | None = None,
        release_type: ReleaseType = ReleaseType.MANUAL,
        earliest_release_date: str | None = None,
    ) -> AppStoreVersion:
        """Create a new version on the app store.

        :param version: The version to create
        :param app_id: The ID of the app
        :param platform: The platform this app is (defaults to iOS)
        :param copyright_text: The copyright string to use
        :param uses_idfa: Set to True if this app uses the advertising ID, false otherwise
        :param release_type: The release type to use (defaults to MANUAL)
        :param earliest_release_date: The earliest date to release this app, required when release_type is SCHEDULED

        :raises TypeError: If the release_type is SCHEDULED but earliest_release_date is not supplied
        :raises AppStoreConnectError: On a failure response

        :returns: An AppStoreVersion
        """

        self.log.info(f"Creating version {version} for {app_id}...")

        attributes: dict[str, Any] = {
            "platform": platform.value,
            "versionString": version,
            "releaseType": release_type.value,
        }

        if release_type == ReleaseType.SCHEDULED:
            if not earliest_release_date:
                raise TypeError(
                    "earliest_release_date must be supplied when release_type is SCHEDULED"
                )
            attributes["earliestReleaseDate"] = earliest_release_date

        if copyright_text:
            attributes["copyright"] = copyright_text

        if uses_idfa is not None:
            attributes["usesIdfa"] = uses_idfa

        data = {
            "data": {
                "attributes": attributes,
                "type": "appStoreVersions",
                "relationships": {
                    "app": {
                        "data": {
                            "type": "apps",
                            "id": app_id,
                        }
                    }
                },
            }
        }

        self.log.debug(f"Creation data: {data}")

        return self.http_client.post(
            endpoint="appStoreVersions",
            data=data,
            data_type=AppStoreVersion,
        )



"""Wrapper around the Apple App Store Connect APIs."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import datetime
import logging
from typing import Any, Iterator, Type

import deserialize
import jwt
import requests

from asconnect.exceptions import AppStoreConnectError

NEW_TOKEN_AGE_IN_MINUTES = 15
MINIMUM_TOKEN_AGE = 5


class HttpClient:
    """Base HTTP client for the ASC API."""

    _key_id: str
    _key_contents: str
    _issuer_id: str
    log: logging.Logger

    _credentials_valid: bool
    _cached_token_info: tuple[str, datetime.datetime] | None

    def __init__(
        self,
        *,
        key_id: str,
        key_contents: str,
        issuer_id: str,
        log: logging.Logger,
    ) -> None:
        """Construct a new client object.

        :param key_id: The ID of your key (can be found in app store connect)
        :param key_contents: The contents of your key
        :param issuer_id: The contents of your key (can be found in app store connect
        :param log: Any base logger to be used (one will be created if not supplied)
        """

        self._key_id = key_id
        self._key_contents = key_contents
        self._issuer_id = issuer_id
        self._credentials_valid = False
        self.log = log.getChild("http")

        self._cached_token_info = None

    @property
    def key_contents(self) -> str:
        """Get key contents.

        :returns: Key contents
        """
        return self._key_contents

    @key_contents.setter
    def key_contents(self, value: str) -> None:
        """Set key contents.

        :param value: The new value to set to
        """
        self._key_contents = value
        self._credentials_valid = False

    @property
    def key_id(self) -> str:
        """Get key ID.

        :returns: Key ID
        """
        return self._key_id

    @key_id.setter
    def key_id(self, value: str) -> None:
        """Set key ID.

        :param value: The new value to set to
        """
        self._key_id = value
        self._credentials_valid = False

    @property
    def issuer_id(self) -> str:
        """Get issuer ID.

        :returns: Issuer ID
        """
        return self._issuer_id

    @issuer_id.setter
    def issuer_id(self, value: str) -> None:
        """Set issuer ID.

        :param value: The new value to set to
        """
        self._issuer_id = value
        self._credentials_valid = False

    def generate_token(self) -> str:
        """Generate a new JWT token.

        :returns: The JWT token as a string
        """

        # Apple state that performance is improved if we re-use the same token
        # instead of generating a new one each time
        if self._cached_token_info is not None:
            cached_token, cached_expiration = self._cached_token_info
            # Only return a token with time remaining on it
            if cached_expiration - datetime.datetime.now() > datetime.timedelta(
                minutes=MINIMUM_TOKEN_AGE
            ):
                self.log.debug(f"Using cached token. Expiration time: {cached_expiration}")
                return cached_token

            self._cached_token_info = None

        # Tokens more than 20 minutes in the future are invalid, so our new one is less than that
        expiration = datetime.datetime.now() + datetime.timedelta(minutes=NEW_TOKEN_AGE_IN_MINUTES)

        # Details at https://developer.apple.com/documentation/appstoreconnectapi/generating_tokens_for_api_requests
        token = jwt.encode(
            {
                "iss": self.issuer_id,
                "exp": int(expiration.timestamp()),
                "aud": "appstoreconnect-v1",
            },
            self.key_contents,
            algorithm="ES256",
            headers={"kid": self.key_id, "typ": "JWT"},
        )

        if not isinstance(token, str):
            token = token.decode("utf-8")

        self.log.debug(f"Generating new token. Expiration: {expiration}")
        self._cached_token_info = (token, expiration)

        return token

    def generate_url(self, endpoint: str) -> str:
        """Generate a URL for an endpoint.

        :param endpoint: The endpoint to generate the URL for

        :returns: An endpoint URL
        """
        _ = self
        return f"https://api.appstoreconnect.apple.com/v1/{endpoint}"

    def verify_response(self, response: requests.Response) -> None:
        """Perform some checks on the response.

        :param response: The response to check
        """
        if response.status_code >= 200 and response.status_code < 300:
            self._credentials_valid = True

    def log_response(self, response: requests.Response) -> None:
        """Log the respose

        :param response: The response to log
        """
        self.log.debug(f"Response: {response.status_code}, {response.text}")

    def get(
        self,
        *,
        data_type: Type,
        endpoint: str | None = None,
        url: str | None = None,
        log_response: bool = False,
        attempts: int = 3,
    ) -> Iterator[Any]:
        """Perform a GET to the endpoint specified.

        Either endpoint or url must be specified. url will take precedence if
        both are specified.

        :param data_type: The class to deserialize the data of the response to
        :param endpoint: The endpoint to perform the GET on
        :param url: The full URL to perform the GET on
        :param log_response: A flag indicates whether to log the response
        :param attempts: Number of attempts remaining to try this call

        :raises ValueError: If neither url or endpoint are specified
        :raises AppStoreConnectError: If an error with the API occurs

        :returns: The raw response
        """
        token = self.generate_token()

        if url is None:
            if endpoint is None:
                raise ValueError("Either `endpoint` or `url` must be set")
            url = self.generate_url(endpoint)

        iterations = 0

        while True:
            iterations += 1

            raw_response = requests.get(
                url,
                headers={"Authorization": f"Bearer {token}"},
            )

            if log_response:
                self.log_response(raw_response)

            self.verify_response(raw_response)

            try:
                response_data = self.extract_data(raw_response)
            except AppStoreConnectError as ex:
                if attempts > 1 and (
                    ex.response.status_code >= 500
                    or (ex.response.status_code == 401 and self._credentials_valid)
                ):
                    yield from self.get(
                        data_type=data_type,
                        endpoint=endpoint,
                        url=url,
                        log_response=log_response,
                        attempts=attempts - 1,
                    )
                    return

                raise

            if response_data["data"] is None:
                yield from []
            else:
                deserialized_data = deserialize.deserialize(data_type, response_data["data"])

                if isinstance(deserialized_data, list):
                    yield from deserialized_data
                else:
                    yield deserialized_data

            if response_data.get("links") is None:
                break

            if response_data["links"].get("next") is None:
                break

            url = response_data["links"]["next"]
            assert url is not None

        # This is only used for debugging, so we just assert so it isn't flagged as unused
        assert iterations != 0

    def patch(
        self,
        *,
        data_type: Type | None = None,
        endpoint: str | None = None,
        url: str | None = None,
        data: Any,
        log_response: bool = False,
    ) -> Any:
        """Perform a PATCH to the endpoint specified.

        Either endpoint or url must be specified. url will take precedence if
        both are specified.

        :param data_type: The class to deserialize the data of the response to
        :param endpoint: The endpoint to perform the GET on
        :param url: The full URL to perform the GET on
        :param data: Some JSON serializable data to send
        :param log_response: A flag indicates whether to log the response

        :raises AppStoreConnectError: If we don't get a 200 response back
        :raises ValueError: If neither url or endpoint are specified

        :returns: The raw response
        """
        token = self.generate_token()

        if url is None:
            if endpoint is None:
                raise ValueError("Either `endpoint` or `url` must be set")
            url = self.generate_url(endpoint)

        raw_response = requests.patch(
            url,
            json=data,
            headers={"Authorization": f"Bearer {token}", "Content-Type": "application/json"},
        )

        if log_response:
            self.log_response(raw_response)

        self.verify_response(raw_response)

        if raw_response.status_code == 204:
            return None

        if raw_response.status_code != 200:
            raise AppStoreConnectError(raw_response)

        response_data = self.extract_data(raw_response)

        if data_type is None:
            return None

        return deserialize.deserialize(data_type, response_data["data"])

    def post(
        self,
        *,
        endpoint: str | None = None,
        url: str | None = None,
        data: Any,
        data_type: Type | None = None,
        log_response: bool = False,
    ) -> Any:
        """Perform a POST to the endpoint specified.

        Either endpoint or url must be specified. url will take precedence if
        both are specified.

        :param endpoint: The endpoint to perform the GET on
        :param url: The full URL to perform the GET on
        :param data: Some JSON serializable data to send
        :param data_type: The data type to deserialize the response to
        :param log_response: A flag indicates whether to log the response

        :raises ValueError: If neither url or endpoint are specified
        :raises AppStoreConnectError: If we get a failure response back from the API

        :returns: The raw response
        """
        token = self.generate_token()

        if url is None:
            if endpoint is None:
                raise ValueError("Either `endpoint` or `url` must be set")
            url = self.generate_url(endpoint)

        raw_response = requests.post(
            url,
            json=data,
            headers={"Authorization": f"Bearer {token}", "Content-Type": "application/json"},
        )

        if log_response:
            self.log_response(raw_response)

        self.verify_response(raw_response)

        if raw_response.status_code == 201:

            if data_type is None:
                return None

            response_data = self.extract_data(raw_response)

            return deserialize.deserialize(data_type, response_data["data"])

        if raw_response.status_code >= 200 and raw_response.status_code < 300:
            return None

        raise AppStoreConnectError(raw_response)

    def delete(
        self,
        *,
        endpoint: str | None = None,
        url: str | None = None,
        log_response: bool = False,
    ) -> requests.Response:
        """Perform a DELETE to the endpoint specified.

        Either endpoint or url must be specified. url will take precedence if
        both are specified.

        :param endpoint: The endpoint to perform the GET on
        :param url: The full URL to perform the GET on
        :param log_response: A flag indicates whether to log the response

        :raises ValueError: If neither url or endpoint are specified

        :returns: The raw response
        """
        token = self.generate_token()

        if url is None:
            if endpoint is None:
                raise ValueError("Either `endpoint` or `url` must be set")
            url = self.generate_url(endpoint)

        raw_response = requests.delete(
            url,
            headers={"Authorization": f"Bearer {token}", "Content-Type": "application/json"},
        )

        if log_response:
            self.log_response(raw_response)

        self.verify_response(raw_response)

        return raw_response

    def put_chunk(
        self,
        *,
        url: str,
        additional_headers: dict[str, str],
        data: bytes,
        use_auth_header: bool = True,
        log_response: bool = False,
    ) -> requests.Response:
        """Perform a PUT to the url specified

        :param url: The full URL to perform the PUT on
        :param additional_headers: The additional headers to add
        :param data: The raw data to upload
        :param use_auth_header: A flag indicates whether an auth header will be included in the request
        :param log_response: A flag indicates whether to log the response

        :returns: The raw response
        """
        token = self.generate_token()

        headers = additional_headers

        if use_auth_header and ("Authorization" not in headers):
            headers["Authorization"] = f"Bearer {token}"

        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/json"

        raw_response = requests.put(url=url, data=data, headers=headers)

        if log_response:
            self.log_response(raw_response)

        self.verify_response(raw_response)

        return raw_response

    def extract_data(self, response: requests.Response) -> Any:
        """Validate a response from the API and extract the data

        :param response: The response to validate

        :raises AppStoreConnectError: On any failure to validate

        :returns: Any data in the response
        """
        _ = self

        if not response.ok:
            raise AppStoreConnectError(response)

        return response.json()



"""User Models for the API"""

import datetime

import deserialize

from asconnect.models.common import BaseAttributes, Links, Relationship, Resource


@deserialize.auto_snake()
@deserialize.parser("created_date", datetime.datetime.fromisoformat)
class CustomerReviewAttributes(BaseAttributes):
    """Represents build attributes."""

    body: str
    created_date: datetime.datetime
    rating: int
    reviewer_nickname: str
    title: str
    territory: str


@deserialize.key("identifier", "id")
class CustomerReview(Resource):
    """Represents a user."""

    identifier: str
    attributes: CustomerReviewAttributes
    relationships: dict[str, Relationship] | None
    links: Links



"""Wrapper for altool."""

import enum
import logging
import subprocess
import time


class Platform(enum.Enum):
    """The platforms that altool supports."""

    IOS = "ios"
    MACOS = "osx"
    TVOS = "appletvos"


def _check_should_restart(line: str) -> bool:
    """Check if the line indicates that the upload should be re-run on failure.

    :param line: The output line to check

    :returns: True if it indicates a restart would help, False otherwise
    """

    for match in [
        "Error: Server returned an invalid MIME type: text/plain",
        "Error: The request timed out.",
        "status code 401, auth issue.",
    ]:
        if match in line:
            return True

    return False


def upload(
    *,
    ipa_path: str,
    platform: Platform,
    key_id: str,
    issuer_id: str,
    log: logging.Logger | None = None,
    attempt: int = 1,
    max_attempts: int = 3,
) -> None:
    """Upload a build to app store connect.

    :param ipa_path: The path to the ipa to upload
    :param platform: The platform the app is for
    :param key_id: The ID for the key used for auth
    :param issuer_id: The ID of the issuer of the auth key
    :param log: Any base logger to be used (one will be created if not supplied)
    :param attempt: The attempt this is
    :param max_attempts: The number of attempts allowed

    :raises CalledProcessError: If something goes wrong during upload
    """

    if log:
        log = log.getChild(__name__)
    else:
        log = logging.getLogger(__name__)

    command = [
        "xcrun",
        "altool",
        "--upload-app",
        "-f",
        ipa_path,
        "-t",
        platform.value,
        "-apiKey",
        key_id,
        "--apiIssuer",
        issuer_id,
        "--verbose",
    ]

    log.info(
        "Beginning upload. This can take a while and will not show any output while in progress."
    )

    upload_process = subprocess.Popen(  # pylint: disable=consider-using-with
        command,
        universal_newlines=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
    )

    assert upload_process.stdout is not None
    command_output = ""
    should_restart = False
    for line in iter(upload_process.stdout.readline, ""):
        log.info(line.rstrip())
        if _check_should_restart(line):
            should_restart = True
        command_output += line
    upload_process.stdout.close()
    upload_process.wait()

    if upload_process.returncode == 0:
        return

    if should_restart and attempt < max_attempts:
        log.info("Upload failed due to intermittent issue. Will sleep for 1 minute and try again.")
        time.sleep(60)
        upload(
            ipa_path=ipa_path,
            platform=platform,
            key_id=key_id,
            issuer_id=issuer_id,
            log=log,
            attempt=attempt + 1,
            max_attempts=max_attempts,
        )
        return

    raise subprocess.CalledProcessError(
        upload_process.returncode,
        command,
        "Failed to upload the build. Please see the logs for more information.",
    )



#!/usr/bin/env python3

"""Wrapper around the Apple App Store Connect APIs."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from asconnect.client import Client
from asconnect.models import App
from asconnect.models import Build

from asconnect.altool import Platform



"""Wrapper around the Apple App Store Connect APIs."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
from typing import Iterator

from asconnect.httpclient import HttpClient
from asconnect.models import AppInfoLocalization, AppInfo, AppStoreVersionLocalization
from asconnect.utilities import update_query_parameters


class AppInfoClient:
    """Wrapper class around the ASC API."""

    log: logging.Logger
    http_client: HttpClient

    def __init__(
        self,
        *,
        http_client: HttpClient,
        log: logging.Logger,
    ) -> None:
        """Construct a new client object.

        :param http_client: The API HTTP client
        :param log: Any base logger to be used (one will be created if not supplied)
        """

        self.http_client = http_client
        self.log = log.getChild("appinfo")

    def get_app_info(self, *, app_id: str) -> list[AppInfo]:
        """Get the app info for an app.

        If there are two, one will be marked as ready for sale, the other will
        be one that is being prepared.

        :param app_id: The app ID to get the info for

        :returns: A list to AppInfoLocalization
        """
        self.log.debug(f"Getting app info for {app_id}")
        url = self.http_client.generate_url(f"apps/{app_id}/appInfos")

        return list(self.http_client.get(url=url, data_type=list[AppInfo]))

    def get_localizations(
        self,
        *,
        app_info_id: str,
        locale: str | None = None,
    ) -> Iterator[AppInfoLocalization]:
        """Get the app info for an app.

        :param app_info_id: The app info ID to get the localized info for
        :param locale: The version to filter on (if any)

        :returns: An iterator to AppInfoLocalization
        """

        self.log.debug(f"Getting localizations for {app_info_id} (locale={locale})")
        url = self.http_client.generate_url(f"appInfos/{app_info_id}/appInfoLocalizations")

        query_parameters = {}

        if locale:
            query_parameters["filter[locale]"] = locale

        url = update_query_parameters(url, query_parameters)

        yield from self.http_client.get(url=url, data_type=list[AppInfoLocalization])

    def set_localization_properties(
        self,
        *,
        localization_id: str,
        name: str | None = None,
        privacy_policy_text: str | None = None,
        privacy_policy_url: str | None = None,
        subtitle: str | None = None,
    ) -> AppInfoLocalization:
        """Set the properties on an app info localization

        Any left as None will be ignored.

        :param localization_id: The ID of the localization to patch
        :param name: The name of the app
        :param privacy_policy_text: The text of the privacy policy
        :param privacy_policy_url: The URL of the privacy policy
        :param subtitle: The sub-title for the app

        :returns: The new updated app info localization
        """

        self.log.info("Setting localization properties")

        attributes = {}

        if name:
            attributes["name"] = name

        if privacy_policy_text:
            attributes["privacyPolicyText"] = privacy_policy_text

        if privacy_policy_url:
            attributes["privacyPolicyUrl"] = privacy_policy_url

        if subtitle:
            attributes["subtitle"] = subtitle

        data = {
            "data": {
                "attributes": attributes,
                "type": "appInfoLocalizations",
                "id": localization_id,
            }
        }

        self.log.debug(f"Localization properties: {data}")

        return self.http_client.patch(
            endpoint=f"appInfoLocalizations/{localization_id}",
            data=data,
            data_type=AppInfoLocalization,
        )

    def set_localization_version_properties(
        self,
        *,
        version_localization_id: str,
        description: str | None = None,
        keywords: str | None = None,
        marketing_url: str | None = None,
        promotional_text: str | None = None,
        support_url: str | None = None,
        whats_new: str | None = None,
    ) -> AppInfoLocalization:
        """Set the properties on an app version localization

        Any left as None will be ignored.

        :param version_localization_id: The ID of the localization to patch
        :param description: The description for the app store
        :param keywords: The keywords for the app (comma separated)
        :param marketing_url: The marketing URL
        :param promotional_text: Any promotional text (can be set after submission)
        :param support_url: The support URL
        :param whats_new: The What's New text for the version (release notes)

        :returns: The new updated app version localization
        """

        self.log.info("Setting localization version properties")

        attributes = {}

        if description:
            attributes["description"] = description

        if keywords:
            attributes["keywords"] = keywords

        if marketing_url:
            attributes["marketingUrl"] = marketing_url

        if promotional_text:
            attributes["promotionalText"] = promotional_text

        if support_url:
            attributes["supportUrl"] = support_url

        if whats_new:
            attributes["whatsNew"] = whats_new

        data = {
            "data": {
                "attributes": attributes,
                "type": "appStoreVersionLocalizations",
                "id": version_localization_id,
            }
        }

        self.log.debug(f"Localization version properties: {data}")

        return self.http_client.patch(
            endpoint=f"appStoreVersionLocalizations/{version_localization_id}",
            data=data,
            data_type=AppStoreVersionLocalization,
        )



"""Sorting options."""

import enum


class BuildsSort(enum.Enum):
    """Orders that builds can be sorted."""

    PRE_RELEASE_VERSION = "preReleaseVersion"
    PRE_RELEASE_VERSION_REVERSED = "-preReleaseVersion"
    UPLOADED_DATE = "uploadedDate"
    UPLOADED_DATE_REVERSED = "-uploadedDate"
    VERSION = "version"
    VERSION_REVERSED = "-version"


class CustomerReviewSort(enum.Enum):
    """Orders that customer reviews can be sorted."""

    RATING_ASC = "rating"
    RATING_DESC = "-rating"
    CREATED_DATE_ASC = "createdDate"
    CREATED_DATE_DESC = "-createdDate"



"""Exceptions"""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from typing import Any

import requests


class AppStoreConnectError(Exception):
    """An error response from the API."""

    identifier: str
    code: str
    status: int
    title: str
    detail: str
    source: Any
    response: requests.Response

    def __init__(self, response: requests.Response):
        """Create a new instance.

        :param response: The HTTP response

        :raises ValueError: If we can't decode the data
        """

        self.response = response

        data = response.json()

        if not isinstance(data, dict):
            raise ValueError(f"Could not decode App Store Connect error: {data}")

        data = data["errors"]

        if not isinstance(data, list):
            raise ValueError(f"Could not decode App Store Connect error: {data}")

        data = data[0]

        self.identifier = data.get("id", "<unknown>")
        self.status = int(data.get("status", -1))
        self.code = data.get("code", "<unknown>")
        self.title = data.get("title", "<unknown>")
        self.detail = data.get("detail", "<unknown>")
        self.source = data.get("source")

        super().__init__(f"[{self.status}] {self.title} ({self.code}): {self.detail}")



"""Wrapper around the Apple App Store Connect APIs."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
import os
from typing import Iterator

from asconnect.exceptions import AppStoreConnectError
from asconnect.httpclient import HttpClient
from asconnect.models import AppScreenshotSet, AppScreenshot, ScreenshotDisplayType, UploadOperation
from asconnect.utilities import md5_file


class ScreenshotClient:
    """Wrapper class around the ASC API."""

    log: logging.Logger
    http_client: HttpClient

    def __init__(
        self,
        *,
        http_client: HttpClient,
        log: logging.Logger,
    ) -> None:
        """Construct a new client object.

        :param http_client: The API HTTP client
        :param log: Any base logger to be used (one will be created if not supplied)
        """

        self.http_client = http_client
        self.log = log.getChild("screenshot")

    def get_sets(
        self,
        *,
        localization_id: str,
    ) -> Iterator[AppScreenshotSet]:
        """Get the screenshot sets for an app localization.

        :param localization_id: The localization ID to get the screenshot sets for

        :returns: An iterator to ScreenshotSet
        """
        self.log.debug(f"Getting screenshot sets for {localization_id}...")
        url = self.http_client.generate_url(
            f"appStoreVersionLocalizations/{localization_id}/appScreenshotSets"
        )
        yield from self.http_client.get(url=url, data_type=list[AppScreenshotSet])

    def delete_set(self, *, screenshot_set_id: str, delete_all_screenshots: bool = True) -> None:
        """Delete a screenshot set.

        :param screenshot_set_id: The ID of the screenshot set to delete
        :param delete_all_screenshots: If this is True, any screenshots will be deleted first.

        :raises AppStoreConnectError: On failure to delete
        """

        self.log.info(f"Deleting screenshot set {screenshot_set_id}")

        if delete_all_screenshots:
            self.delete_screenshots_in_set(screenshot_set_id=screenshot_set_id)

        url = self.http_client.generate_url(f"appScreenshotSets/{screenshot_set_id}")
        raw_response = self.http_client.delete(url=url)

        if raw_response.status_code != 204:
            raise AppStoreConnectError(raw_response)

    def get_screenshots(
        self,
        *,
        screenshot_set_id: str,
    ) -> Iterator[AppScreenshot]:
        """Get the screenshots for a set.

        :param screenshot_set_id: The screenshot set ID to get the screenshots for

        :returns: An iterator to AppScreenshot
        """
        self.log.debug(f"Getting screenshots {screenshot_set_id}")
        url = self.http_client.generate_url(f"appScreenshotSets/{screenshot_set_id}/appScreenshots")
        yield from self.http_client.get(url=url, data_type=list[AppScreenshot])

    def delete_screenshot(self, *, screenshot_id: str) -> None:
        """Delete a screenshot.

        :param screenshot_id: The ID of the screenshot to delete

        :raises AppStoreConnectError: On failure to delete
        """
        self.log.info(f"Deleting screenshot {screenshot_id}")
        url = self.http_client.generate_url(f"appScreenshots/{screenshot_id}")
        raw_response = self.http_client.delete(url=url)

        if raw_response.status_code != 204:
            raise AppStoreConnectError(raw_response)

    def delete_screenshots_in_set(self, *, screenshot_set_id: str) -> None:
        """Delete all screenshots in set.

        :param screenshot_set_id: The set to delete the screenshots in
        """
        self.log.info(f"Deleting screenshots in set {screenshot_set_id}")
        for screenshot in self.get_screenshots(screenshot_set_id=screenshot_set_id):
            self.log.info(f"Deleting screenshot {screenshot.attributes.file_name}")
            self.delete_screenshot(screenshot_id=screenshot.identifier)

    def delete_all_sets_in_localization(self, *, localization_id: str) -> None:
        """Delete all the sets in a localization.

        :param localization_id: The localization to delete the sets from
        """
        self.log.info(f"Deleting all screenshot sets in localization {localization_id}")
        for screenshot_set in self.get_sets(localization_id=localization_id):
            self.log.info(
                f"Deleting screenshot set: {screenshot_set.attributes.screenshot_display_type.value}"
            )
            self.delete_set(
                screenshot_set_id=screenshot_set.identifier, delete_all_screenshots=True
            )

    def create_set(
        self, *, localization_id: str, display_type: ScreenshotDisplayType
    ) -> AppScreenshotSet:
        """Create a screenshot set for an app localization.

        :param localization_id: The localization ID to create the screenshot set for
        :param display_type: The type of preview that the set is for

        :raises AppStoreConnectError: On error when creating the set

        :returns: The new screenshot set
        """

        self.log.info(f"Creating screenshot set {localization_id} / {display_type}")

        data = {
            "data": {
                "attributes": {"screenshotDisplayType": display_type.value},
                "type": "appScreenshotSets",
                "relationships": {
                    "appStoreVersionLocalization": {
                        "data": {
                            "type": "appStoreVersionLocalizations",
                            "id": localization_id,
                        }
                    }
                },
            }
        }

        self.log.debug(f"Data: {data}")

        return self.http_client.post(
            endpoint="appScreenshotSets",
            data=data,
            data_type=AppScreenshotSet,
        )

    def _create_screenshot_reservation(
        self, *, file_path: str, screenshot_set_id: str
    ) -> AppScreenshot:
        """Create a screenshot reservation

        :param file_path: The path to the screenshot to reserve
        :param screenshot_set_id: The id for the screenshot set to reserve in

        :raises AppStoreConnectError: On error when creating the set

        :returns: The new screenshot set
        """

        self.log.debug(f"Creating screenshot reservarion for {screenshot_set_id} at {file_path}")

        file_name = os.path.basename(file_path)
        file_size = os.path.getsize(file_path)

        return self.http_client.post(
            endpoint="appScreenshots",
            data={
                "data": {
                    "attributes": {"fileName": file_name, "fileSize": file_size},
                    "type": "appScreenshots",
                    "relationships": {
                        "appScreenshotSet": {
                            "data": {
                                "type": "appScreenshotSets",
                                "id": screenshot_set_id,
                            }
                        }
                    },
                }
            },
            data_type=AppScreenshot,
        )

    def _upload_screenshot_contents(
        self, *, file_path: str, upload_operations: list[UploadOperation], use_auth_header: bool
    ) -> None:
        """Upload a screenshots contents

        :param file_path: The path to the screenshot to upload
        :param upload_operations: The upload operations for the screenshot
        :param use_auth_header: If this is true, an auth header will be included in the upload request

        :raises AppStoreConnectError: On error when creating the set
        """

        self.log.debug(
            f"Uploading screenshot contents {file_path}: {upload_operations}, use auth header: {use_auth_header}"
        )

        # Start by ordering the upload oeprations by offset (so we can just go in order)
        upload_operations = sorted(upload_operations, key=lambda operation: operation.offset)

        with open(file_path, "rb") as screenshot:
            for operation in upload_operations:
                data = screenshot.read(operation.length)
                headers = {header.name: header.value for header in operation.request_headers}
                raw_response = self.http_client.put_chunk(
                    url=operation.url,
                    additional_headers=headers,
                    data=data,
                    use_auth_header=use_auth_header,
                    log_response=True,
                )
                # TODO Check this
                assert raw_response.ok

    def _set_screenshot_uploaded(
        self, *, screenshot: AppScreenshot, file_hash: str
    ) -> AppScreenshot:
        """Marks a screenshot as uploaded

        :param screenshot: The screenshot to mark as uploaded
        :param file_hash: The MD5 of the file

        :returns: The new screenshot
        """

        self.log.debug(f"Setting screenshot uploaded {screenshot}: {file_hash}")

        return self.http_client.patch(
            endpoint=f"appScreenshots/{screenshot.identifier}",
            data={
                "data": {
                    "attributes": {"uploaded": True, "sourceFileChecksum": file_hash},
                    "type": "appScreenshots",
                    "id": screenshot.identifier,
                }
            },
            data_type=AppScreenshot,
        )

    def upload_screenshot(
        self, *, file_path: str, screenshot_set_id: str, use_auth_header: bool = True
    ) -> AppScreenshot:
        """Upload a screenshot

        :param file_path: The path to the screenshot to upload
        :param screenshot_set_id: The id for the screenshot set to upload to
        :param use_auth_header: If this is true, an auth header will be included in the upload request

        :return: The screenshot
        """

        self.log.info(f"Uploading screenshot {file_path} to set {screenshot_set_id}")

        checksum = md5_file(file_path)

        screenshot = self._create_screenshot_reservation(
            file_path=file_path, screenshot_set_id=screenshot_set_id
        )

        assert screenshot.attributes.upload_operations is not None

        self._upload_screenshot_contents(
            file_path=file_path,
            upload_operations=screenshot.attributes.upload_operations,
            use_auth_header=use_auth_header,
        )

        return self._set_screenshot_uploaded(screenshot=screenshot, file_hash=checksum)



"""Wrapper around the Apple App Store Connect APIs."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
import os
import time
from typing import Iterator

from asconnect.httpclient import HttpClient

from asconnect.altool import upload, Platform
from asconnect.models import App, Build, BuildBetaDetail
from asconnect.sorting import BuildsSort
from asconnect.utilities import update_query_parameters, next_or_none, write_key


class BuildClient:
    """Wrapper class around the ASC API."""

    log: logging.Logger
    http_client: HttpClient

    def __init__(
        self,
        *,
        http_client: HttpClient,
        log: logging.Logger,
    ) -> None:
        """Construct a new client object.

        :param http_client: The API HTTP client
        :param log: Any base logger to be used (one will be created if not supplied)
        """

        self.http_client = http_client
        self.log = log.getChild("build")

    def get_builds(
        self,
        *,
        url: str | None = None,
        sort: BuildsSort | None = None,
        build_number: str | None = None,
        version: str | None = None,
        app_id: str | None = None,
    ) -> Iterator[Build]:
        """Get all builds.

        :param url: The URL to use (will be generated if not supplied)
        :param sort: The sort option to use
        :param build_number: Filter to just this build number
        :param version: Filter to just this version
        :param app_id: Filter to just this app

        :returns: A list of builds
        """

        self.log.info("Getting builds...")

        if url is None:
            url = self.http_client.generate_url("builds")

        query_parameters = {}

        if sort:
            query_parameters["sort"] = sort.value

        if build_number:
            query_parameters["filter[version]"] = build_number

        if app_id:
            query_parameters["filter[app]"] = app_id

        if version:
            query_parameters["filter[preReleaseVersion.version]"] = version

        url = update_query_parameters(url, query_parameters)

        yield from self.http_client.get(url=url, data_type=list[Build])

    def get_build_from_identifier(self, identifier: str) -> Build | None:
        """Get a build from its identifier

        :param identifier: The unique identifier for the build (_not_ the build number)

        :returns: A build if found, None otherwise
        """

        self.log.info(f"Getting build {identifier}")

        url = self.http_client.generate_url(f"builds/{identifier}")

        return next_or_none(self.http_client.get(url=url, data_type=Build))

    def get_from_build_number(self, bundle_id: str, build_number: str) -> Build | None:
        """Get a build from its build number.

        :param bundle_id: The bundle ID of the app
        :param build_number: The build number for the build to get

        :returns: The build if found, None otherwise
        """

        self.log.info(f"Getting build from build number {build_number} for bundle {bundle_id}")

        for build in self.get_builds(build_number=build_number):
            self.log.debug(f"Checking build {build}")

            # TODO use app id directly for this
            assert build.relationships is not None
            related_link = build.relationships["app"].links.related
            assert related_link is not None
            app = next_or_none(self.http_client.get(url=related_link, data_type=App))

            if not app:
                break

            if app.bundle_id == bundle_id:
                return build

        return None

    def wait_for_build_to_process(
        self, bundle_id: str, build_number: str, wait_time: int = 30
    ) -> Build:
        """Wait for a build to finish processing.

        :param bundle_id: The ID of the app
        :param build_number: The build number for the build to wait for
        :param wait_time: The time to wait between checks for processing completion in seconds

        :returns: The build when finished processing
        """
        build = None

        while True:
            self.log.info("Waiting for build to appear...")
            build = self.get_from_build_number(bundle_id, build_number)
            if build is not None:
                break
            time.sleep(wait_time)

        if build.attributes.processing_state != "PROCESSING":
            return build

        while True:
            build = self.get_build_from_identifier(build.identifier)
            assert build is not None

            if build.attributes.processing_state != "PROCESSING":
                return build

            self.log.info(
                f"Build {build_number} has not finished processing. Will check again in {wait_time} seconds..."
            )
            time.sleep(wait_time)

    def get_beta_detail(self, build: Build) -> BuildBetaDetail | None:
        """Get the build beta details.

        :param build: The build to get the beta details for

        :returns: An iterator to the beta app localizations
        """
        self.log.debug(f"Get beta detail for build {build}")
        assert build.relationships is not None
        url = build.relationships["buildBetaDetail"].links.related
        return next_or_none(self.http_client.get(url=url, data_type=BuildBetaDetail))

    def upload(self, ipa_path: str, platform: Platform, max_attempts: int = 3) -> None:
        """Upload a build to App Store Connect.

        :param ipa_path: The path to the IPA
        :param platform: The platform the app is for
        :param max_attempts: The number of attempts allowed
        """

        self.log.info(f"Uploading IPA {ipa_path} for platform {platform}")

        key_path = write_key(self.http_client.key_id, self.http_client.key_contents)

        try:
            upload(
                ipa_path=ipa_path,
                platform=platform,
                key_id=self.http_client.key_id,
                issuer_id=self.http_client.issuer_id,
                log=self.log,
                max_attempts=max_attempts,
            )

        finally:
            os.remove(key_path)



"""Wrapper around the Apple App Store Connect APIs."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging


from asconnect.httpclient import HttpClient
from asconnect.app_client import AppClient
from asconnect.app_info_client import AppInfoClient
from asconnect.beta_review_client import BetaReviewClient
from asconnect.build_client import BuildClient
from asconnect.reviews_client import ReviewsClient
from asconnect.screenshot_client import ScreenshotClient
from asconnect.users_client import UsersClient
from asconnect.version_client import VersionClient

# pylint: disable=too-many-public-methods


class Client:
    """Wrapper class around the ASC API."""

    log: logging.Logger
    http_client: HttpClient

    app: AppClient
    app_info: AppInfoClient
    beta_review: BetaReviewClient
    build: BuildClient
    reviews: ReviewsClient
    screenshots: ScreenshotClient
    users: UsersClient
    version: VersionClient

    def __init__(
        self,
        *,
        key_id: str,
        key_contents: str,
        issuer_id: str,
        log: logging.Logger | None = None,
    ) -> None:
        """Construct a new client object.

        :param key_id: The ID of your key (can be found in app store connect)
        :param key_contents: The contents of your key
        :param issuer_id: The contents of your key (can be found in app store connect
        :param log: Any base logger to be used (one will be created if not supplied)
        """

        if log is None:
            self.log = logging.getLogger("asconnect")
        else:
            self.log = log.getChild("asconnect")

        self.http_client = HttpClient(
            key_id=key_id, key_contents=key_contents, issuer_id=issuer_id, log=self.log
        )

        self.app = AppClient(http_client=self.http_client, log=self.log)
        self.app_info = AppInfoClient(http_client=self.http_client, log=self.log)
        self.beta_review = BetaReviewClient(http_client=self.http_client, log=self.log)
        self.build = BuildClient(http_client=self.http_client, log=self.log)
        self.reviews = ReviewsClient(http_client=self.http_client, log=self.log)
        self.screenshots = ScreenshotClient(http_client=self.http_client, log=self.log)
        self.users = UsersClient(http_client=self.http_client, log=self.log)
        self.version = VersionClient(http_client=self.http_client, log=self.log)
